
/*
 **********************************************************************
 *     cardwo.c - PCM output HAL for emu10k1 driver
 *     Copyright 1999, 2000 Creative Labs, Inc.
 *
 **********************************************************************
 *
 *     Date                 Author          Summary of changes
 *     ----                 ------          ------------------
 *     October 20, 1999     Bertrand Lee    base code release
 *
 **********************************************************************
 *
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 *     USA.
 *
 **********************************************************************
 */

#include <linux/poll.h>
#include "hwaccess.h"
#include "8010.h"
#include "voicemgr.h"
#include "cardwo.h"
#include "audio.h"

/* Volume calcs */

static int set_volume_instance(struct emu10k1_waveout *card_waveout, struct wave_out *wave_out, struct voice_param *left)
{
	/* only applicable for playback */
	u32 volL, volR, vol = 0;

	volL = (wave_out->localvol & 0xffff);
	volR = ((wave_out->localvol >> 16) & 0xffff);

	if (wave_out->globalvolFactor) {
		volL = ((u32) (((u16) card_waveout->globalvol & 0xffff) * (u16) volL)) / 0xffff;
		volR = ((u32) (((u16) (card_waveout->globalvol >> 16) & 0xffff) * ((u16) volR))) / 0xffff;
	}

	/* BIG ASSUMPTION HERE THAT DEFAULT WAVE PAN/AUX IS 0xff/0xff */
	/* New volume and pan */

	if (volL == volR) {
		vol = volL;
		left->send_c = 0xff;
		left->send_b = 0xff;
	} else {
		if (volL > volR) {
			vol = volL;
			left->send_c = 0xff;
			left->send_b = (char) ((volR * 255) / vol);
		} else {
			vol = volR;
			left->send_b = 0xff;
			left->send_c = (char) ((volL * 255) / vol);
		}
	}

	left->initial_attn = 0xff & sumVolumeToAttenuation(vol * 2);

	return vol;
}

static void query_format(struct wave_format *wave_fmt)
{
	if ((wave_fmt->channels != 1) && (wave_fmt->channels != 2))
		wave_fmt->channels = 2;

	if (wave_fmt->samplingrate >= 0x2EE00)
		wave_fmt->samplingrate = 0x2EE00;

	if ((wave_fmt->bitsperchannel != 8) && (wave_fmt->bitsperchannel != 16))
		wave_fmt->bitsperchannel = 16;

	if (wave_fmt->bitsperchannel == 16)
		wave_fmt->is_16bit = 1;
	else
		wave_fmt->is_16bit = 0;

	if (wave_fmt->channels == 1)
		wave_fmt->is_stereo = 0;
	else
		wave_fmt->is_stereo = 1;

	wave_fmt->bytespersample = (wave_fmt->is_stereo + 1) * (wave_fmt->is_16bit + 1);

	wave_fmt->bytespersec = wave_fmt->bytespersample * wave_fmt->samplingrate;

	return;
}

static int alloc_buffer(struct emu10k1_card *card, struct waveout_buffer *buffer)
{
	u32 pageindex, pagecount;
	unsigned long busaddx;
	int i;

	buffer->pos = 0;
	buffer->silence_pos = 0;
	buffer->hw_pos = 0;

	DPD(2, "requested pages is: %d\n", buffer->pages);

	if ((buffer->emupageindex = emu10k1_addxmgr_alloc(buffer->pages * PAGE_SIZE, card)) < 0)
		return CTSTATUS_ERROR;

	/* Fill in virtual memory table */
	for (pagecount = 0; pagecount < buffer->pages; pagecount++) {
		if ((buffer->addr[pagecount] = pci_alloc_consistent(card->pci_dev, PAGE_SIZE, &buffer->dma_handle[pagecount])) == NULL) {
			buffer->pages = pagecount;
			return CTSTATUS_ERROR;
		}

		DPD(2, "Virtual Addx: %p\n", buffer->addr[pagecount]);

		for (i = 0; i < PAGE_SIZE / EMUPAGESIZE; i++) {
			busaddx = buffer->dma_handle[pagecount] + i * EMUPAGESIZE;

			DPD(3, "Bus Addx: %lx\n", busaddx);

			pageindex = buffer->emupageindex + pagecount * PAGE_SIZE / EMUPAGESIZE + i;

			((u32 *) card->virtualpagetable.addr)[pageindex] = (busaddx * 2) | pageindex;
		}
	}

	return CTSTATUS_SUCCESS;
}

static void free_buffer(struct emu10k1_card *card, struct waveout_buffer *buffer)
{
	u32 pagecount, pageindex;
	int i;

	if (buffer->emupageindex < 0)
		return;

	for (pagecount = 0; pagecount < buffer->pages; pagecount++) {
		pci_free_consistent(card->pci_dev, PAGE_SIZE, buffer->addr[pagecount], buffer->dma_handle[pagecount]);

		for (i = 0; i < PAGE_SIZE / EMUPAGESIZE; i++) {
			pageindex = buffer->emupageindex + pagecount * PAGE_SIZE / EMUPAGESIZE + i;
			((u32 *) card->virtualpagetable.addr)[pageindex] = (card->silentpage.dma_handle * 2) | pageindex;
		}
	}

	emu10k1_addxmgr_free(card, buffer->emupageindex);
	buffer->emupageindex = -1;

	return;
}

static int get_voice(struct emu10k1_card *card, struct woinst *woinst)
{
	struct emu10k1_waveout *card_waveout = card->waveout;
	struct wave_out *wave_out = &woinst->wave_out;
	struct voice_allocdesc voice_allocdesc;
	struct voice_param *left, *right;
	u32 size;

	/* Allocate voices here, if no voices available, return error.
	 * Init voice_allocdesc first.*/

	voice_allocdesc.usage = VOICE_USAGE_PLAYBACK;

	voice_allocdesc.flags = 0;

	if (woinst->device == 1)
		voice_allocdesc.flags |= VOICE_FLAGS_FXRT2;

	if (woinst->format.is_stereo)
		voice_allocdesc.flags |= VOICE_FLAGS_STEREO;

	if (woinst->format.is_16bit)
		voice_allocdesc.flags |= VOICE_FLAGS_16BIT;

	if (emu10k1_voice_alloc(card, &woinst->voice, &voice_allocdesc) != CTSTATUS_SUCCESS)
		return CTSTATUS_ERROR;

	/* voice initialization */

	left = &woinst->voice.params;

	/* Calculate pitch */
	left->initial_pitch = (u16) (srToPitch(woinst->format.samplingrate) >> 8);

	DPD(2, "Initial pitch --> %x\n", left->initial_pitch);

	/* Easy way out.. gotta calculate value */
	left->pitch_target = 0;
	left->volume_target = 0;
	left->FC_target = 0;

	left->byampl_env_sustain = 0x7f;
	left->byampl_env_decay = 0x7f;

	if (wave_out->globalreverbFactor) {
		u8 t = (card_waveout->globalreverb & 0xff) + (wave_out->localreverb & 0xff);

		left->send_a = (t > 255) ? 255 : t;
	} else {
		left->send_a = 0;
	}

	if (wave_out->globalchorusFactor) {
		u8 t = (card_waveout->globalchorus & 0xff) + (wave_out->localchorus & 0xff);

		left->send_d = (t > 255) ? 255 : t;
	} else {
		left->send_d = 0;
	}

	set_volume_instance(card_waveout, wave_out, left);

	left->pan_target = left->send_c;
	left->aux_target = left->send_b;

	size = woinst->buffer.size / woinst->format.bytespersample;
	left->start = 2 * (woinst->buffer.emupageindex << 11) / woinst->format.bytespersample;
	left->end = left->start + size;
	left->startloop = left->start;
	left->endloop = left->end;

	if (woinst->voice.linked_voice) {
		right = &woinst->voice.linked_voice->params;

		right->initial_pitch = left->initial_pitch;

		/* Easy way out.. gotta calculate value */
		right->pitch_target = 0;
		right->volume_target = 0;
		right->FC_target = 0;

		right->byampl_env_sustain = 0x7f;
		right->byampl_env_decay = 0x7f;

		right->send_d = left->send_d;
		right->send_a = left->send_a;

		/* Left output of right channel is always zero */
		right->send_c = 0;

		/* Update right channel aux */
		right->pan_target = 0;
		right->send_b = left->send_b;
		right->aux_target = right->send_b;

		/* Zero out right output of left channel */
		left->send_b = 0;
		left->aux_target = 0;

		/* Update right channel attenuation */
		right->initial_attn = left->initial_attn;

		right->start = left->start;
		right->end = left->end;
		right->startloop = left->startloop;
		right->endloop = left->endloop;

	}

	DPD(2, "voice: start=%x, end=%x, startloop=%x, endloop=%x\n", left->start, left->end, left->startloop, left->endloop);

	emu10k1_voice_playback_setup(&woinst->voice);

	return CTSTATUS_SUCCESS;
}

int emu10k1_waveout_open(struct emu10k1_wavedevice *wave_dev)
{
	struct emu10k1_card *card = wave_dev->card;
	struct woinst *woinst = wave_dev->woinst;
	struct wave_out *wave_out = &woinst->wave_out;
	u32 delay;

	DPF(2, "emu10k1_waveout_open()\n");

	/* Assign default local volume */
	wave_out->localvol = 0xffffffff;
	wave_out->localreverb = 0xffffffff;
	wave_out->localchorus = 0xffffffff;
	wave_out->globalvolFactor = 0xffff;
	wave_out->globalreverbFactor = 0xffff;
	wave_out->globalchorusFactor = 0xffff;

	if (alloc_buffer(card, &woinst->buffer) != CTSTATUS_SUCCESS) {
		ERROR();
		emu10k1_waveout_close(wave_dev);
		return CTSTATUS_ERROR;
	}

	woinst->buffer.fill_silence = 0;
	woinst->buffer.silence_bytes = 0;

	if (get_voice(card, woinst) != CTSTATUS_SUCCESS) {
		ERROR();
		emu10k1_waveout_close(wave_dev);
		return CTSTATUS_ERROR;
	}

	delay = (48000 * woinst->buffer.fragment_size) / woinst->format.bytespersec;

	emu10k1_timer_install(card, &woinst->timer, emu10k1_waveout_bh, (unsigned long) wave_dev, delay / 2);

	woinst->state = WAVE_STATE_OPEN;

	return CTSTATUS_SUCCESS;
}

void emu10k1_waveout_close(struct emu10k1_wavedevice *wave_dev)
{
	struct emu10k1_card *card = wave_dev->card;
	struct woinst *woinst = wave_dev->woinst;

	DPF(2, "emu10k1_waveout_close()\n");

	emu10k1_waveout_stop(wave_dev);

	emu10k1_timer_uninstall(card, &woinst->timer);

	emu10k1_voice_free(&woinst->voice);

	free_buffer(card, &woinst->buffer);

	woinst->state = WAVE_STATE_CLOSED;

	return;
}

void emu10k1_waveout_start(struct emu10k1_wavedevice *wave_dev)
{
	struct emu10k1_card *card = wave_dev->card;
	struct woinst *woinst = wave_dev->woinst;

	DPF(2, "emu10k1_waveout_start()\n");
	/* Actual start */
	emu10k1_voice_start(&woinst->voice);

	emu10k1_timer_enable(card, &woinst->timer);

	woinst->state |= WAVE_STATE_STARTED;

	return;
}

int emu10k1_waveout_setformat(struct emu10k1_wavedevice *wave_dev, struct wave_format *format)
{
	struct emu10k1_card *card = wave_dev->card;
	struct woinst *woinst = wave_dev->woinst;
	u32 delay;

	DPF(2, "emu10k1_waveout_setformat()\n");

	if (woinst->state & WAVE_STATE_STARTED)
		return CTSTATUS_ERROR;

	query_format(format);

	if (woinst->format.samplingrate != format->samplingrate ||
	    woinst->format.channels != format->channels ||
	    woinst->format.bitsperchannel != format->bitsperchannel) {

		woinst->format = *format;

		if (woinst->state == WAVE_STATE_CLOSED)
			return CTSTATUS_SUCCESS;

		emu10k1_timer_uninstall(card, &woinst->timer);
		emu10k1_voice_free(&woinst->voice);

		if (get_voice(card, woinst) != CTSTATUS_SUCCESS) {
			ERROR();
			emu10k1_waveout_close(wave_dev);
			return CTSTATUS_ERROR;
		}

		delay = (48000 * woinst->buffer.fragment_size) / woinst->format.bytespersec;

		emu10k1_timer_install(card, &woinst->timer, emu10k1_waveout_bh, (unsigned long) wave_dev, delay / 2);
	}

	return CTSTATUS_SUCCESS;
}

void emu10k1_waveout_stop(struct emu10k1_wavedevice *wave_dev)
{
	struct emu10k1_card *card = wave_dev->card;
	struct woinst *woinst = wave_dev->woinst;

	DPF(2, "emu10k1_waveout_stop()\n");

	if (!(woinst->state & WAVE_STATE_STARTED))
		return;

	emu10k1_timer_disable(card, &woinst->timer);

	/* Stop actual voice */
	emu10k1_voice_stop(&woinst->voice);

	emu10k1_waveout_update(woinst);

	DPD(2, "position is %x\n", woinst->buffer.hw_pos / woinst->format.bytespersample);

	woinst->state &= ~WAVE_STATE_STARTED;

	return;
}

void emu10k1_waveout_getxfersize(struct woinst *woinst, u32 * size)
{
	struct waveout_buffer *buffer = &woinst->buffer;
	u32 pending;

	if (buffer->hw_pos > buffer->silence_pos) {
		*size = buffer->hw_pos - buffer->silence_pos;
	} else {
		*size = buffer->size + buffer->hw_pos - buffer->silence_pos;
	}

	pending = buffer->size - *size;

	buffer->fill_silence = (pending < buffer->fragment_size) ? 1 : 0;

	if (pending > buffer->silence_bytes) {
		*size += buffer->silence_bytes;
	} else {
		*size = buffer->size;
		buffer->pos = buffer->hw_pos;
	}

	return;
}

static void copy_block(u32 dst, u8 * src, u32 len, void **pt)
{
	int i, j, k;

	i = dst / PAGE_SIZE;
	j = dst % PAGE_SIZE;
	k = (len > PAGE_SIZE - j) ? PAGE_SIZE - j : len;
	copy_from_user(pt[i] + j, src, k);
	len -= k;
	while (len >= PAGE_SIZE) {
		copy_from_user(pt[++i], src + k, PAGE_SIZE);
		k += PAGE_SIZE;
		len -= PAGE_SIZE;
	}
	copy_from_user(pt[++i], src + k, len);

	return;
}

static void fill_block(u32 dst, u8 val, u32 len, void **pt)
{
	int i, j, k;

	i = dst / PAGE_SIZE;
	j = dst % PAGE_SIZE;
	k = (len > PAGE_SIZE - j) ? PAGE_SIZE - j : len;
	memset(pt[i] + j, val, k);
	len -= k;
	while (len >= PAGE_SIZE) {
		memset(pt[++i], val, PAGE_SIZE);
		len -= PAGE_SIZE;
	}
	memset(pt[++i], val, len);

	return;
}

void emu10k1_waveout_xferdata(struct woinst *woinst, u8 * data, u32 * size)
{
	struct waveout_buffer *buffer = &woinst->buffer;
	u32 sizetocopy, sizetocopy_now, start;
	unsigned long flags;

	sizetocopy = min(buffer->size, *size);
	*size = sizetocopy;

	if (!sizetocopy)
		return;

	spin_lock_irqsave(&woinst->lock, flags);
	buffer->silence_bytes = 0;
	sizetocopy_now = buffer->size - buffer->pos;

	start = buffer->pos;

	if (sizetocopy > sizetocopy_now) {
		sizetocopy -= sizetocopy_now;
		buffer->pos = sizetocopy;
		buffer->silence_pos = buffer->pos;
		spin_unlock_irqrestore(&woinst->lock, flags);

		copy_block(start, data, sizetocopy_now, buffer->addr);
		copy_block(0, data + sizetocopy_now, sizetocopy, buffer->addr);
	} else {
		if (sizetocopy == sizetocopy_now)
			buffer->pos = 0;
		else
			buffer->pos += sizetocopy;

		buffer->silence_pos = buffer->pos;
		spin_unlock_irqrestore(&woinst->lock, flags);

		copy_block(start, data, sizetocopy, buffer->addr);
	}

	return;
}

void emu10k1_waveout_fillsilence(struct woinst *woinst)
{
	struct waveout_buffer *buffer = &woinst->buffer;
	u16 filldata;
	u32 sizetocopy, sizetocopy_now, start;
	unsigned long flags;

	sizetocopy = woinst->buffer.fragment_size;

	if (woinst->format.is_16bit)
		filldata = 0x0000;
	else
		filldata = 0x8080;

	spin_lock_irqsave(&woinst->lock, flags);
	buffer->silence_bytes += sizetocopy;
	start = buffer->silence_pos;
	sizetocopy_now = buffer->size - start;

	if (sizetocopy > sizetocopy_now) {
		sizetocopy -= sizetocopy_now;
		buffer->silence_pos = sizetocopy;
		spin_unlock_irqrestore(&woinst->lock, flags);
		fill_block(start, filldata, sizetocopy_now, buffer->addr);
		fill_block(0, filldata, sizetocopy, buffer->addr);
	} else {
		if (sizetocopy == sizetocopy_now)
			buffer->silence_pos = 0;
		else
			buffer->silence_pos += sizetocopy;

		spin_unlock_irqrestore(&woinst->lock, flags);

		fill_block(start, filldata, sizetocopy, buffer->addr);
	}

	return;
}

void emu10k1_waveout_update(struct woinst *woinst)
{
	u32 hw_pos;

	/* There is no actual start yet */
	if (!(woinst->state & WAVE_STATE_STARTED)) {
		hw_pos = woinst->buffer.hw_pos;
	} else {
		u32 samples;

		emu10k1_voice_getcontrol(&woinst->voice, CCCA_CURRADDR, &hw_pos);

		hw_pos -= woinst->voice.params.start;

		/* Get number of bytes in play buffer per channel.
		 * If 8 bit mode is enabled, this needs to be changed. */
		hw_pos *= woinst->format.bytespersample;

		/* Refer to voicemgr.c, CA is not started at zero.
		 * We need to take this into account. */
		samples = 64 * (woinst->format.is_stereo + 1) - 4 * (woinst->format.is_16bit + 1);

		if (hw_pos >= samples)
			hw_pos -= samples;
		else
			hw_pos += woinst->buffer.size - samples;
	}

	if (hw_pos >= woinst->buffer.hw_pos)
		woinst->total_played += hw_pos - woinst->buffer.hw_pos;
	else
		woinst->total_played += woinst->buffer.size + hw_pos - woinst->buffer.hw_pos;

	woinst->buffer.hw_pos = hw_pos;

	return;
}
