
/*
 **********************************************************************
 *     voicemgr.c - Voice manager for emu10k1 driver
 *     Copyright 1999, 2000 Creative Labs, Inc.
 *
 **********************************************************************
 *
 *     Date                 Author          Summary of changes
 *     ----                 ------          ------------------
 *     October 20, 1999     Bertrand Lee    base code release
 *
 **********************************************************************
 *
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 *     USA.
 *
 **********************************************************************
 */

#include "voicemgr.h"
#include "8010.h"

int emu10k1_voice_alloc(struct emu10k1_card *card, struct emu_voice *voice, struct voice_allocdesc *voiceallocdesc)
{
	u8 *voicetable = card->voicetable;
	int i;
	unsigned long flags;

	DPF(2, "emu10k1_voice_alloc()\n");

	if (voiceallocdesc->flags & VOICE_FLAGS_STEREO) {
		spin_lock_irqsave(&card->lock, flags);
		for (i = 0; i < NUM_G; i += 2)
			if ((voicetable[i] == VOICE_USAGE_FREE) && (voicetable[i + 1] == VOICE_USAGE_FREE)) {
				voicetable[i] = voiceallocdesc->usage;
				voicetable[i + 1] = voiceallocdesc->usage;
				break;
			}
		spin_unlock_irqrestore(&card->lock, flags);

		if (i >= NUM_G)
			return CTSTATUS_ERROR;

		if ((voice->linked_voice = (struct emu_voice *) kmalloc(sizeof(struct emu_voice), GFP_KERNEL)) == NULL) {
			spin_lock_irqsave(&card->lock, flags);
			voicetable[i] = VOICE_USAGE_FREE;
			voicetable[i + 1] = VOICE_USAGE_FREE;
			spin_unlock_irqrestore(&card->lock, flags);
			return CTSTATUS_ERROR;
		}

		voice->card = card;
		voice->flags = VOICE_FLAGS_VOICEMASTER | voiceallocdesc->flags;
		voice->usage = voiceallocdesc->usage;
		voice->num = i;

		voice->linked_voice->card = card;
		voice->linked_voice->flags = VOICE_FLAGS_STEREOSLAVE | voiceallocdesc->flags;
		voice->linked_voice->usage = voiceallocdesc->usage;
		voice->linked_voice->num = i + 1;
		voice->linked_voice->linked_voice = NULL;
	} else {
		spin_lock_irqsave(&card->lock, flags);
		for (i = 0; i < NUM_G; i++)
			if (voicetable[i] == VOICE_USAGE_FREE) {
				voicetable[i] = voiceallocdesc->usage;
				break;
			}
		spin_unlock_irqrestore(&card->lock, flags);

		if (i >= NUM_G)
			return CTSTATUS_ERROR;

		voice->card = card;
		voice->flags = VOICE_FLAGS_VOICEMASTER | voiceallocdesc->flags;
		voice->usage = voiceallocdesc->usage;
		voice->num = i;
		voice->linked_voice = NULL;
	}

	while (voice != NULL) {

		DPD(2, " voice allocated -> %d\n", voice->num);

		sblive_writeptr(card, IFATN, voice->num, 0xffff);
		sblive_writeptr(card, DCYSUSV, voice->num, 0);
		sblive_writeptr(card, VTFT, voice->num, 0x0000ffff);
		sblive_writeptr(card, PTRX, voice->num, 0);

		voice = voice->linked_voice;
	}

	return CTSTATUS_SUCCESS;
}

void emu10k1_voice_free(struct emu_voice *voice)
{
	struct emu10k1_card *card = voice->card;
	struct emu_voice *voice_tmp = voice;
	unsigned dcysusv;
	u32 cra, sample;
	unsigned long flags;

	DPF(2, "emu10k1_voice_free()\n");

	if (voice->usage == VOICE_USAGE_FREE)
		return;

	voice_tmp = voice;

	while (voice != NULL) {

		DPD(2, " voice freed -> %d\n", voice->num);

		dcysusv = sblive_readptr(card, DCYSUSV, voice->num) & (DCYSUSV_PHASE1_MASK | DCYSUSV_SUSTAINLEVEL_MASK | DCYSUSV_DECAYTIME_MASK);
		sblive_writeptr(card, DCYSUSV, voice->num, dcysusv | 0x00);

		sblive_writeptr(card, VTFT, voice->num, 0x0000ffff);
		sblive_writeptr(card, PTRX_PITCHTARGET, voice->num, 0);
		sblive_writeptr(card, CVCF, voice->num, 0x0000ffff);
		sblive_writeptr(card, CPF, voice->num, 0);

		sample = (voice->flags & VOICE_FLAGS_16BIT) ? 0 : 0x80808080;
		cra = sblive_readptr(card, CCR, voice->num) & CCR_READADDRESS_MASK;
		sblive_writeptr(card, CCR, voice->num, cra);
		cra = (cra >> 18) & 0xf;
		sblive_writeptr(card, CD0 + cra, voice->num, sample);
		cra = (cra + 0x1) & 0xf;
		sblive_writeptr(card, CD0 + cra, voice->num, sample);

		voice->usage = VOICE_USAGE_FREE;
		voice = voice->linked_voice;
	}

	spin_lock_irqsave(&card->lock, flags);
	card->voicetable[voice_tmp->num] = VOICE_USAGE_FREE;
	if (voice_tmp->linked_voice) {
		card->voicetable[voice_tmp->num + 1] = VOICE_USAGE_FREE;
		kfree(voice_tmp->linked_voice);
	}
	spin_unlock_irqrestore(&card->lock, flags);

	return;
}

/*       Sets up a voices for Wave Playback */

void emu10k1_voice_playback_setup(struct emu_voice *voice)
{
	struct emu10k1_card *card = voice->card;
	u32 sample, cra = 0, start = 0;

	DPF(2, "emu10k1_voice_playback_setup()\n");

	while (voice != NULL) {
		sblive_writeptr(card, FXRT, voice->num, (voice->flags & VOICE_FLAGS_FXRT2) ? 0xd23c0000 : 0xd01c0000);

		/* Stop CA */
		/* Assumption that PT is alreadt 0 so no harm overwriting */
		sblive_writeptr(card, PTRX, voice->num, (voice->params.send_a << 8) | voice->params.send_b);

		if (voice->flags & VOICE_FLAGS_VOICEMASTER) {
			if (voice->linked_voice != NULL) {
				/* Set stereo bit */
				cra = 64;
				sblive_writeptr(card, CPF, voice->num, CPF_STEREO_MASK);
				sblive_writeptr(card, CPF, voice->num + 1, CPF_STEREO_MASK);
			} else {
				cra = 32;
				sblive_writeptr(card, CPF, voice->num, 0);
			}

			if (voice->flags & VOICE_FLAGS_16BIT)
				sample = 0;
			else {
				cra = cra * 2;
				sample = 0x80808080;
			}
			cra -= 4;

			if (voice->linked_voice != NULL) {
				/* CCR_READADDRESS_MASK */
				sblive_writeptr(card, CCR, voice->num, 0x3c << 16);
				sblive_writeptr(card, CCR, voice->num + 1, cra << 16);
				sblive_writeptr(card, CDE, voice->num + 1, sample);
				sblive_writeptr(card, CDF, voice->num + 1, sample);
				start = voice->params.start + cra / 2;
			} else {
				sblive_writeptr(card, CCR, voice->num, 0x1c << 16);	/* FIXME: Is 0x1c correct? */
				sblive_writeptr(card, CDE, voice->num, sample);
				sblive_writeptr(card, CDF, voice->num, sample);
				start = voice->params.start + cra;
			}

			if (start > voice->params.endloop) {
				start -= voice->params.endloop;

				if (voice->linked_voice != NULL)
					cra = (cra << 25) | 0x1bc0000 | ((cra - start) << 9);
				else
					cra = (cra << 25) | 0x11c0000 | ((cra - start) << 9);

				start += voice->params.startloop;

				if (start >= voice->params.endloop)
					start = voice->params.endloop - 1;
			} else if (voice->linked_voice != NULL)
				cra = (cra << 25) | (0x3c << 16);
			else
				cra = (cra << 25) | (0x1c << 16);

			start |= CCCA_INTERPROM_0;
		}

		/* CSL, ST, CA */
		sblive_writeptr(card, DSL, voice->num, voice->params.endloop | (voice->params.send_d << 24));
		sblive_writeptr(card, PSST, voice->num, voice->params.startloop | (voice->params.send_c << 24));

		if (voice->flags & VOICE_FLAGS_16BIT)
			sblive_writeptr(card, CCCA, voice->num, start);
		else
			sblive_writeptr(card, CCCA, voice->num, start | CCCA_8BITSELECT);

		/* Clear filter delay memory */
		sblive_writeptr(card, Z1, voice->num, 0);
		sblive_writeptr(card, Z2, voice->num, 0);

		/* Invalidate maps */
		sblive_writeptr(card, MAPA, voice->num, MAP_PTI_MASK | (card->silentpage.dma_handle * 2));
		sblive_writeptr(card, MAPB, voice->num, MAP_PTI_MASK | (card->silentpage.dma_handle * 2));

		/* Fill cache */
		if (voice->flags & VOICE_FLAGS_VOICEMASTER)
			sblive_writeptr(card, CCR, voice->num, cra);

		sblive_writeptr(card, CVCF, voice->num, 0x0000ffff);
		sblive_writeptr(card, VTFT, voice->num, 0x0000ffff);

		/* modulation envelope */
		sblive_writeptr(card, ATKHLDM, voice->num, 0);
		sblive_writeptr(card, DCYSUSM, voice->num, 0x007f);
		sblive_writeptr(card, LFOVAL1, voice->num, 0x8000);
                sblive_writeptr(card, LFOVAL2, voice->num, 0x8000);
		sblive_writeptr(card, FMMOD, voice->num, 0);
                sblive_writeptr(card, TREMFRQ, voice->num, 0);
                sblive_writeptr(card, FM2FRQ2, voice->num, 0);
		sblive_writeptr(card, ENVVAL, voice->num, 0xbfff);

		/* volume envelope */
		sblive_writeptr(card, ATKHLDV, voice->num, 0x7f7f);
		sblive_writeptr(card, ENVVOL, voice->num, 0xbfff);

		/* filter envelope */
		sblive_writeptr(card, PEFE_FILTERAMOUNT, voice->num, 0x7f);

		/* pitch envelope */
		sblive_writeptr(card, PEFE_PITCHAMOUNT, voice->num, 0);


#ifdef PRIVATE_PCM_VOLUME
		{
			int i;

			for (i = 0; i < MAX_PCM_CHANNELS; i++) {
				if (sblive_pcm_volume[i].channel_l == voice->num) {
					voice->params.initial_attn = (sblive_pcm_volume[i].channel_r < NUM_G) ? sblive_pcm_volume[i].attn_l :
					    // test for mono channel (reverse logic is correct here!)
					    (sblive_pcm_volume[i].attn_r >
					     sblive_pcm_volume[i].attn_l) ? sblive_pcm_volume[i].attn_l : sblive_pcm_volume[i].attn_r;
					DPD(2, "set left volume  %d\n", voice->params.initial_attn);
					break;
				} else if (sblive_pcm_volume[i].channel_r == voice->num) {
					voice->params.initial_attn = sblive_pcm_volume[i].attn_r;
					DPD(2, "set right volume  %d\n", voice->params.initial_attn);
					break;
				}
			}
		}
#endif

		voice->params.FC_target = 0xffff;
		voice->params.pitch_target = (u16) (IP_TO_CP(voice->params.initial_pitch) >> 16);

		voice = voice->linked_voice;
	}

	return;
}

void emu10k1_voice_start(struct emu_voice *voice)
{
	struct emu10k1_card *card = voice->card;

	DPF(2, "emu10k1_voice_start()\n");

	while (voice != NULL) {
		sblive_writeptr(card, IFATN, voice->num, 0xff00 | voice->params.initial_attn);

		sblive_writeptr(card, VTFT, voice->num, ((u32) voice->params.volume_target << 16)
				| voice->params.FC_target);
		sblive_writeptr(card, CVCF, voice->num, ((u32) voice->params.volume_target << 16)
				| voice->params.FC_target);
		sblive_writeptr(card, DCYSUSV, voice->num, (voice->params.byampl_env_sustain << 8)
				| 0x80 | voice->params.byampl_env_decay);

		/* Using StopOnLoop for MIDI stops the playback
		   too early, which may cause a DC level to be played
		   until the note is released. */

		if (voice->usage == VOICE_USAGE_MIDI)
			emu10k1_clear_stop_on_loop(card, voice->num);
		else {
			if (voice->params.startloop > voice->params.end)
				emu10k1_set_stop_on_loop(card, voice->num);
			else
				emu10k1_clear_stop_on_loop(card, voice->num);
		}

		sblive_writeptr(card, PTRX_PITCHTARGET, voice->num, voice->params.pitch_target);
		if (voice->flags & VOICE_FLAGS_VOICEMASTER)
                        sblive_writeptr(card, CPF_CURRENTPITCH, voice->num, voice->params.pitch_target);

		sblive_writeptr(card, IP, voice->num, voice->params.initial_pitch);

		voice = voice->linked_voice;
	}

	return;
}

void emu10k1_voice_stop(struct emu_voice *voice)
{
	struct emu10k1_card *card = voice->card;

	DPF(2, "emu10k1_voice_stop()\n");

	while (voice != NULL) {
		sblive_writeptr(card, IFATN, voice->num, 0xffff);
		sblive_writeptr(card, VTFT, voice->num, 0x0000ffff);
		sblive_writeptr(card, PTRX_PITCHTARGET, voice->num, 0);
		sblive_writeptr(card, IP, voice->num, 0);

		voice = voice->linked_voice;
	}

	return;
}

void emu10k1_voice_setcontrol(struct emu_voice *voice, struct voice_cntlset *setting, u32 numparam)
{
	struct emu10k1_card *card = voice->card;
	int count;

	for (count = 0; count < numparam; count++)
		sblive_writeptr(card, setting[count].paramID, voice->num, setting[count].value);

	return;
}

void emu10k1_voice_getcontrol(struct emu_voice *voice, u32 controlid, u32 * value)
{
	struct emu10k1_card *card = voice->card;

	*value = sblive_readptr(card, controlid, voice->num);

	return;
}
